import 'package:dynamic_widget/dynamic_widget.dart';
import 'package:flutter/material.dart';
import 'package:flutter_form_builder/flutter_form_builder.dart';

import '../layout_form_renderer.dart';
import '../util.dart';

enum TxtFieldInputType { TEXT, EMAIL, NUM, PASSWORD, NAME, PHONE, CURRENCY }

extension TxtFieldInputTypeExtension on TxtFieldInputType {
  String get stringType{
    switch (this) {
      case TxtFieldInputType.TEXT:
        return "Text";
      case TxtFieldInputType.EMAIL:
        return "Email";
      case TxtFieldInputType.NUM:
        return "Number";
      case TxtFieldInputType.PASSWORD:
        return "Password";
      case TxtFieldInputType.NAME:
        return "Name";
      case TxtFieldInputType.PHONE:
        return "Phone";
      case TxtFieldInputType.CURRENCY:
        return "Currency";
      default:
        return "Text";
    }
  }
  int get type {
    switch (this) {
      case TxtFieldInputType.TEXT:
        return 0;
      case TxtFieldInputType.EMAIL:
        return 1;
      case TxtFieldInputType.NUM:
        return 2;
      case TxtFieldInputType.PASSWORD:
        return 3;
      case TxtFieldInputType.NAME:
        return 4;
      case TxtFieldInputType.PHONE:
        return 5;
      case TxtFieldInputType.CURRENCY:
        return 6;
      default:
        return 0;
    }
  }
}

class TxtField extends StatefulWidget {
  final Map<String, dynamic> fieldProp;
  final ClickListener? listener;

  TxtField({Key? key, required this.fieldProp, required this.listener})
      : super(key: key);

  @override
  _TxtFieldState createState() => _TxtFieldState();
}

class _TxtFieldState extends State<TxtField> {
  bool _gotData = false;
  bool _hideData = false;
  String _prevInputStr = "";
  late FocusNode txtFocusNode;

  @override
  void initState() {
    _prevInputStr = LayoutForm.getFormFieldInitValue(widget.fieldProp) ?? "";
    _gotData = _prevInputStr != "";
    _hideData = widget.fieldProp[LayoutForm.PROP_INPUT_TYPE]?.toString() == TxtFieldInputType.PASSWORD.type.toString() || widget.fieldProp[LayoutForm.PROP_INPUT_TYPE]?.toString() == TxtFieldInputType.PASSWORD.stringType;
    super.initState();
    txtFocusNode = FocusNode();
    txtFocusNode.addListener(() {
      if (txtFocusNode.hasFocus) {
        print("Gain focus");
      } else {
        print("Lost focus");
      }
    });
  }

  @override
  void dispose() {
    // Clean up the focus node when the Form is disposed.
    txtFocusNode.dispose();
    super.dispose();
  }

  Widget _buildSuffixIcon() {
    Map<String, dynamic> fp = widget.fieldProp;
    String inputType = LayoutForm.getFormFieldPropValue(context, fp, LayoutForm.PROP_INPUT_TYPE, false).toString();
    bool? enabled = LayoutForm.getFormFieldPropValue(
        context, fp, LayoutForm.PROP_INPUT_ENABLED, false);

    Icon icon = Icon(Icons.highlight_remove_rounded, color: Colors.red);
    if (inputType == TxtFieldInputType.PASSWORD.type.toString() || inputType == TxtFieldInputType.PASSWORD.stringType) {
      icon = Icon(_hideData ? Icons.visibility : Icons.visibility_off);
    } else if (!_gotData) {
      return Opacity(opacity: 0, child: Icon(Icons.ac_unit));
    }

    if (!(enabled == null ? true : enabled)) {
      return Opacity(opacity: 0, child: Icon(Icons.ac_unit));
    }

    return InkWell(
        onTap: () {
          if (inputType == TxtFieldInputType.PASSWORD.type.toString() || inputType == TxtFieldInputType.PASSWORD.stringType) {
            setState(() {
              _hideData = !_hideData;
            });
          } else {
            _prevInputStr = "";
            LayoutForm.setFormFieldValue(widget.fieldProp, "");
          }
        },
        child: icon);
  }

  @override
  Widget build(BuildContext context) {
    Map<String, dynamic> fp = widget.fieldProp;
    int? maxLength = LayoutForm.getFormFieldPropValue(context, fp, LayoutForm.PROP_MAX_LENGTH, true);
    maxLength = maxLength == 0 ? null : maxLength;
    bool? enabled = LayoutForm.getFormFieldPropValue(context, fp, LayoutForm.PROP_INPUT_ENABLED, false);
    String? label = LayoutForm.getFormFieldPropValue(context, fp, LayoutForm.PROP_INPUT_LABEL, false);
    String? hint = LayoutForm.getFormFieldPropValue(context, fp, LayoutForm.PROP_INPUT_HINT, false);
    String inputType = LayoutForm.getFormFieldPropValue(context, fp, LayoutForm.PROP_INPUT_TYPE, false).toString();

    return Theme(
        data: new ThemeData(
          primaryColor: HexColor.fromHex("#9599B6"),
          primaryColorDark: HexColor.fromHex("#9599B6"),
        ),
        child: Stack(children: [
          Container(/*color: Colors.yellow,*/ height: 65),
          FormBuilderTextField(
              style: TextStyle(
                  fontSize: 14.0, fontFamily: 'Lato', color: Colors.black),
              name: fp[LayoutForm.PROP_INPUT_ID],
              enabled: enabled != null ? enabled : true,
              autovalidateMode: AutovalidateMode.onUserInteraction,
              maxLength: maxLength,
              decoration: InputDecoration(
                  isDense: true,
                  errorBorder: new OutlineInputBorder(
                      borderSide: new BorderSide(color: Colors.red)),
                  border: new OutlineInputBorder(
                      borderSide:
                          new BorderSide(color: HexColor.fromHex("#9599B6"))),
                  labelText: label != null ? label : "",
                  hintText: hint != null ? hint : "",
                  suffixIcon: _buildSuffixIcon()),
              validator: (val) {
                return _validate(val);
              },
              keyboardType: (inputType == TxtFieldInputType.NUM.type.toString() || inputType == TxtFieldInputType.NUM.stringType)
                  ? TextInputType.number
                  : TextInputType.text,
              textInputAction: TextInputAction.next,
              onChanged: (str) {
                String editedTxt = str;
                bool validInput = true;
                if (editedTxt != null && editedTxt.length > 0) {
                  if (inputType == TxtFieldInputType.NUM.type ||
                      inputType == TxtFieldInputType.CURRENCY.type) {
                    if (inputType == TxtFieldInputType.NUM.type) {
                      validInput = editedTxt.isValidNum();
                    }

                    if (inputType == TxtFieldInputType.CURRENCY.type) {
                      validInput = editedTxt.isValidCurrency();
                    }
                  }
                }
                if (validInput) {
                  _prevInputStr = editedTxt;
                  setState(() {
                    _gotData = (editedTxt).length > 0;
                  });
                } else {
                  LayoutForm.setFormFieldValue(fp, _prevInputStr);
                }
              },
              focusNode: txtFocusNode,
              obscureText: _hideData)
        ]));
  }

  String? _validate(dynamic? value) {
    Map<String, dynamic> fp = widget.fieldProp;
    int? minLength = LayoutForm.getFormFieldPropValue(
        context, fp, LayoutForm.PROP_MIN_LENGTH, true);
    String? minLengthErrMsg = LayoutForm.getFormFieldPropValue(
        context, fp, LayoutForm.PROP_MIN_LENGTH_ERR_MSG, true);
    String? emailErrMsg = LayoutForm.getFormFieldPropValue(
        context, fp, LayoutForm.PROP_EMAIL_ERR_MSG, true);
    String? nameErrMsg = LayoutForm.getFormFieldPropValue(
        context, fp, LayoutForm.PROP_NAME_ERR_MSG, true);
    String? phoneErrMsg = LayoutForm.getFormFieldPropValue(
        context, fp, LayoutForm.PROP_PHONE_ERR_MSG, true);
    String? regExpErrMsg = LayoutForm.getFormFieldPropValue(
        context, fp, LayoutForm.PROP_REG_EXP_ERR_MSG, true);
    String? numErrMsg = LayoutForm.getFormFieldPropValue(
        context, fp, LayoutForm.PROP_NUM_ERR_MSG, true);
    String? currencyErrMsg = LayoutForm.getFormFieldPropValue(
        context, fp, LayoutForm.PROP_CURRENCY_ERR_MSG, true);
    String? regExp = LayoutForm.getFormFieldPropValue(
        context, fp, LayoutForm.PROP_REG_EXP, true);
    bool compulsory = LayoutForm.getFormFieldPropValue(
        context, fp, LayoutForm.PROP_COMPULSORY, true);
    String compulsoryErrMsg = LayoutForm.getFormFieldPropValue(
        context, fp, LayoutForm.PROP_COMPULSORY_ERR_MSG, true);

    bool? enabled = LayoutForm.getFormFieldPropValue(
        context, fp, LayoutForm.PROP_INPUT_ENABLED, false);
    String inputType = LayoutForm.getFormFieldPropValue(context, fp, LayoutForm.PROP_INPUT_TYPE, false).toString();

    double? minVal;
    try {
      minVal = LayoutForm.getFormFieldPropValue(
          context, fp, LayoutForm.PROP_MIN_VAL, true);
    } catch (ex) {
      int? intVal = LayoutForm.getFormFieldPropValue(
          context, fp, LayoutForm.PROP_MIN_VAL, true);
      minVal = intVal?.toDouble();
    }

    String? minValErrMsg = LayoutForm.getFormFieldPropValue(
        context, fp, LayoutForm.PROP_MIN_VAL_ERR_MSG, true);

    double? maxVal;
    try {
      maxVal = LayoutForm.getFormFieldPropValue(
          context, fp, LayoutForm.PROP_MAX_VAL, true);
    } catch (ex) {
      int? intVal = LayoutForm.getFormFieldPropValue(
          context, fp, LayoutForm.PROP_MAX_VAL, true);
      maxVal = intVal?.toDouble();
    }

    String? maxValErrMsg = LayoutForm.getFormFieldPropValue(
        context, fp, LayoutForm.PROP_MAX_VAL_ERR_MSG, true);

    if (enabled ?? true) {
      if (inputType == TxtFieldInputType.TEXT.type.toString() ||
          inputType == TxtFieldInputType.PASSWORD.type.toString() ||
          inputType == TxtFieldInputType.NAME.type.toString() ||
          inputType == TxtFieldInputType.EMAIL.type.toString() ||
          inputType == TxtFieldInputType.PHONE.type.toString() ||
          inputType == TxtFieldInputType.TEXT.stringType ||
          inputType == TxtFieldInputType.PASSWORD.stringType ||
          inputType == TxtFieldInputType.NAME.stringType ||
          inputType == TxtFieldInputType.EMAIL.stringType ||
          inputType == TxtFieldInputType.PHONE.stringType
      ) {
        String? val = value;
        if (compulsory) {
          if (val == null || val.length == 0) {
            return compulsoryErrMsg;
          }
        }

        if (minLength != null && (val?.length ?? 0) < minLength) {
          return minLengthErrMsg;
        }

        if ((val?.length ?? 0) > 0) {
          if (inputType == TxtFieldInputType.EMAIL.type.toString() || inputType == TxtFieldInputType.EMAIL.stringType) {
            if (!(val?.isValidEmail() ?? true)) {
              return emailErrMsg;
            }
          }

          if (inputType == TxtFieldInputType.NAME.type.toString() || inputType == TxtFieldInputType.NAME.stringType) {
            if (!(val?.isValidName() ?? true)) {
              return nameErrMsg;
            }
          }

          if (inputType == TxtFieldInputType.PHONE.type.toString() || inputType == TxtFieldInputType.PHONE.stringType) {
            if (!(val?.isValidPhone() ?? true)) {
              return phoneErrMsg;
            }
          }

          if (regExp != null &&
              regExp != "" &&
              (!(val?.isValid(regExp) ?? true))) {
            return regExpErrMsg;
          }
        }
      } else if (inputType == TxtFieldInputType.NUM.type.toString() || inputType == TxtFieldInputType.NUM.stringType ||
          inputType == TxtFieldInputType.CURRENCY.type.toString() || inputType == TxtFieldInputType.CURRENCY.stringType) {
        String? val = value;
        if (compulsory) {
          if (val == null || val.length == 0) {
            return compulsoryErrMsg;
          }
        }

        if ((val?.length ?? 0) > 0) {
          if (inputType == TxtFieldInputType.NUM.type.toString() || inputType == TxtFieldInputType.NUM.stringType) {
            if (!(val?.isValidNum() ?? true)) {
              return numErrMsg;
            }
          }

          if (inputType == TxtFieldInputType.CURRENCY.type.toString() || inputType == TxtFieldInputType.CURRENCY.stringType) {
            if (!(val?.isValidCurrency() ?? true)) {
              return currencyErrMsg;
            }
          }
          double dbl = double.parse(val!);

          if (minVal != null && dbl < minVal) {
            return minValErrMsg;
          }

          if (maxVal != null && dbl > maxVal) {
            return maxValErrMsg;
          }
        }
      }
    }
    return null;
  }
}

extension StrValidator on String {
  bool isValidEmail() {
    return RegExp(
            r'^(([^<>()[\]\\.,;:\s@\"]+(\.[^<>()[\]\\.,;:\s@\"]+)*)|(\".+\"))@((\[[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}\])|(([a-zA-Z\-0-9]+\.)+[a-zA-Z]{2,}))$')
        .hasMatch(this);
  }

  bool isValidName() {
    return RegExp(r"^\s*([A-Za-z]{1,}([\.,] |[-']| ))+[A-Za-z]+\.?\s*$")
        .hasMatch(this);
  }

  bool isValidNum() {
    return RegExp(r"^[0-9]$").hasMatch(this);
  }

  bool isValidCurrency() {
    return RegExp(r"(?=.*?\d)^\$?(([1-9]\d{0,2}(,\d{3})*)|\d+)?(\.\d{1,2})?$")
        .hasMatch(this);
  }

  bool isValidPhone() {
    return RegExp(r"^(?:[+0]9)?[0-9]{10}$").hasMatch(this);
  }

  bool isValid(String regExp) {
    return RegExp(r"" + regExp).hasMatch(this);
  }
}
