import 'package:flutter/material.dart';
import 'package:dynamic_widget/dynamic_widget.dart';

import 'package:flutter/cupertino.dart';
import 'package:flutter_form_builder/flutter_form_builder.dart';

import 'dart:async';
import 'package:http/http.dart' as http;
import 'package:provider/provider.dart';
import 'parser/full_size_column_parser.dart';
import 'parser/partial_size_column_parser.dart';
import 'parser/text_field_parser.dart';
import 'parser/flexible_parser.dart';
import 'parser/card_parser.dart';
import 'widget_prop_change_notifier.dart';

class FormSource {
  final String? url;
  final String? jsonStr;

  FormSource({this.url, this.jsonStr});
}

class PageJsonFormData {
  final String pageId;
  final String pageJson;
  final Map<String, dynamic> formData;

  PageJsonFormData(
      {required this.pageId, required this.pageJson, required this.formData});

  static Future<PageJsonFormData> from(
      FormSource source, Function genInitFormData, String pageId) async {
    String finalJsonStr = "";
    if ((source.jsonStr?.length ?? 0) > 0) {
      finalJsonStr = source.jsonStr!;
    } else if ((source.url?.length ?? 0) > 0) {
      final response =
          await http.get(Uri.parse(source.url!.replaceAll('#pageId#', pageId)));

      if (response.statusCode >= 200 && response.statusCode < 300) {
        finalJsonStr = response.body;
        // finalJsonStr = '''
        // ''';
        print("JSON: $finalJsonStr");
      } else {
        throw Exception(
            'Failed to load json data, use to populate the FORM UI.');
      }
    } else {
      throw Exception('Please specify the form source');
    }
    return PageJsonFormData(
        pageId: pageId,
        pageJson: finalJsonStr,
        formData: await genInitFormData());
  }
}

class LayoutForm extends StatefulWidget {
  static const String PROP_INPUT_TYPE = 'inputType';
  static const String PROP_FORM_ID = 'formId';
  static const String PROP_INPUT_ID = 'inputId';
  static const String PROP_INPUT_LABEL = 'label';
  static const String PROP_INPUT_HINT = 'hint';
  static const String PROP_INPUT_ENABLED = 'enabled';
  static const String PROP_INPUT_VISIBLE = 'visible';
  static const String PROP_COLOR = 'color';
  static const String PROP_OPTIONS = 'options';
  static const String PROP_ELEVATION = 'elevation';
  static const String PROP_ROUNDED_RECTANGLE_BORDER_RADIUS =
      'roundedRectangleBorderRadius';
  static const String PROP_FLEX = 'flex';
  static const String PROP_CHILD = 'child';
  static const String PROP_VALIDATORS = 'validators';
  static const String PROP_MAX_LENGTH = 'maxLength';
  static const String PROP_MIN_LENGTH = 'minLength';
  static const String PROP_MIN_LENGTH_ERR_MSG = 'minLengthErrMsg';
  static const String PROP_COMPULSORY = 'compulsory';
  static const String PROP_COMPULSORY_ERR_MSG = 'compulsoryErrMsg';
  static const String PROP_REG_EXP = 'regExp';
  static const String PROP_REG_EXP_ERR_MSG = 'regExpErrMsg';
  static const String PROP_EMAIL_ERR_MSG = 'emailErrMsg';
  static const String PROP_NAME_ERR_MSG = 'nameErrMsg';
  static const String PROP_PHONE_ERR_MSG = 'phoneErrMsg';
  static const String PROP_NUM_ERR_MSG = 'numErrMsg';
  static const String PROP_CURRENCY_ERR_MSG = 'currencyErrMsg';
  static const String PROP_MAX_VAL = 'maxVal';
  static const String PROP_MAX_VAL_ERR_MSG = 'maxValErrMsg';
  static const String PROP_MIN_VAL = 'minVal';
  static const String PROP_MIN_VAL_ERR_MSG = 'minValErrMsg';

  final ClickListener clickListener;
  final PageJsonFormData pageJsonFormData;
  static Map<String, GlobalKey<FormBuilderState>> allFormKeys =
      Map<String, GlobalKey<FormBuilderState>>();

  LayoutForm(
      {Key? key, required this.clickListener, required this.pageJsonFormData})
      : super(key: key) {
    _addAllCustomParser();
    if (!LayoutForm.allFormKeys.containsKey(pageJsonFormData.pageId)) {
      LayoutForm.allFormKeys[pageJsonFormData.pageId] =
          GlobalKey<FormBuilderState>();
    }
  }

  @override
  _LayoutFormState createState() => _LayoutFormState();

  void _addAllCustomParser() {
    DynamicWidgetBuilder.addParser(CardParser());
    DynamicWidgetBuilder.addParser(TextFieldParser());
    DynamicWidgetBuilder.addParser(FlexibleParser());
    DynamicWidgetBuilder.addParser(FullSizeColumnParser());
    DynamicWidgetBuilder.addParser(PartialSizeColumnParser());
  }

  static dynamic? getFormFieldPropValue(BuildContext context,
      Map<String, dynamic> fp, String name, bool fromValidator) {
    Map<String, dynamic>? widgetChgProps = context
        .watch<WidgetPropChangeNotifier>()
        .props[fp[LayoutForm.PROP_INPUT_ID]];

    bool widgetChgPropsIsNull = false;
    if (widgetChgProps == null) {
      widgetChgPropsIsNull = true;
      widgetChgProps = fp;
    }

    if (fromValidator) {
      if (widgetChgProps.containsKey(LayoutForm.PROP_VALIDATORS)) {
        if (widgetChgProps[LayoutForm.PROP_VALIDATORS][name] != null) {
          return widgetChgProps[LayoutForm.PROP_VALIDATORS][name];
        }
      }
    } else {
      if (widgetChgProps[name] != null) {
        return widgetChgProps[name];
      }
    }

    if (!widgetChgPropsIsNull) {
      if (fromValidator) {
        if (fp.containsKey(LayoutForm.PROP_VALIDATORS)) {
          if (fp[LayoutForm.PROP_VALIDATORS][name] != null) {
            return fp[LayoutForm.PROP_VALIDATORS][name];
          }
        }
      } else {
        if (fp[name] != null) {
          return fp[name];
        }
      }
    }

    return null;
  }

  static FormBuilderFieldState? lookupFormField(
      Map<String, dynamic> mapFromParser) {
    return LayoutForm.allFormKeys[mapFromParser[PROP_FORM_ID]]?.currentState
        ?.fields[mapFromParser[PROP_INPUT_ID]];
  }

  static Map<String, dynamic>? getFormValues(String formKey) {
    final key = LayoutForm.allFormKeys[formKey];
    if (key != null) {
      if (key.currentState != null) {
        final st = key.currentState!;
        st.save();
        if (st.validate()) {
          return st.value;
        }
      }
    }
    return null;
  }

  static void setFormFieldValue(
      Map<String, dynamic> mapFromParser, dynamic value) {
    lookupFormField(mapFromParser)?.didChange(value);
  }

  static void setFormFieldValueWithInputId(
      String inputId, String formId, dynamic value) {
    LayoutForm.allFormKeys[formId]?.currentState?.fields[inputId]
        ?.didChange(value);
  }

  static dynamic getFormFieldInitValue(Map<String, dynamic> mapFromParser) {
    lookupFormField(mapFromParser)?.initialValue;
  }

  static dynamic getFormFieldEditedValue(Map<String, dynamic> mapFromParser) {
    lookupFormField(mapFromParser)?.value;
  }
}

class _LayoutFormState extends State<LayoutForm> {
  @override
  Widget build(BuildContext context) {
    return MultiProvider(
        providers: [
          ChangeNotifierProvider(create: (_) => WidgetPropChangeNotifier())
        ],
        child: FormBuilder(
          skipDisabled: true,
          key: LayoutForm.allFormKeys[widget.pageJsonFormData.pageId],
          initialValue: widget.pageJsonFormData.formData,
          autovalidateMode: AutovalidateMode.always,
          child: _LayoutFormRenderer(
              clickListener: widget.clickListener,
              jsonStr: widget.pageJsonFormData.pageJson),
        ));
  }
}

class _LayoutFormRenderer extends StatelessWidget {
  final String jsonStr;
  final ClickListener clickListener;

  _LayoutFormRenderer(
      {Key? key, required this.clickListener, required this.jsonStr})
      : super(key: key);

  @override
  Widget build(BuildContext context) {
    return FutureBuilder<Widget?>(
        future: _buildWidget(context),
        builder: (BuildContext context, AsyncSnapshot<Widget?> snapshot) {
          if (snapshot.hasError) {
            return Center(
                child:
                    Text('Failed to render the UI due to ${snapshot.error}'));
          }
          return snapshot.hasData
              ? snapshot.data ?? Container()
              : Center(
                  child: Column(
                  mainAxisSize: MainAxisSize.min,
                  children: [
                    CircularProgressIndicator(),
                    SizedBox(height: 5),
                    Text('Rendering...')
                  ],
                ));
        });
  }

  Future<Widget?> _buildWidget(BuildContext context) async {
    try {
      print("LayoutFormRenderer: Rendering dynamic widget with json : $jsonStr");
      return DynamicWidgetBuilder.build(jsonStr, context, clickListener);
    } catch (e, stacktrace) {
      print("Stack Trace: $stacktrace");
      throw e;
    }
  }
}
